@@ -5,6 +5,7 @@ from math import floor
 from time import sleep
 
 import pandas as pd
+import pygsheets
 import requests
 
 module_logger = logging.getLogger(__name__)
@@ -14,8 +15,8 @@ def retry(worker_method, *args, **kwargs):
     """Allows you to retry a function/method three times to overcome network jitters
 
     Retries worker_method three times (for a total of four tries, including the initial attempt), pausing 2^trycount
-    seconds between each retry. Any arguments for worker_method can be passed in as additional parameters to _retry()
-    following worker_method: _retry(foo_method, arg1, arg2, keyword_arg=3)
+    seconds between each retry. Any arguments for worker_method can be passed in as additional parameters to retry()
+    following worker_method: retry(foo_method, arg1, arg2, keyword_arg=3)
 
     Args:
         worker_method (callable): The name of the method to be retried (minus the calling parens)
@@ -106,7 +107,7 @@ def check_fields_match(featurelayer, new_dataframe):
         RuntimeError: If new data contains a field not present in the live data
     """
 
-    live_fields = {field['name'] for field in featurelayer.properties['fields']}
+    live_fields = {field['name'] for field in featurelayer.properties.fields}
     new_fields = set(new_dataframe.columns)
     #: Remove SHAPE field from set (live "featurelayer.properties['fields']" does not expose the 'SHAPE' field)
     try:
@@ -136,11 +137,34 @@ def check_index_column_in_feature_layer(featurelayer, index_column):
         RuntimeError: If index_column is not in featurelayer's fields
     """
 
-    featurelayer_fields = [field['name'] for field in featurelayer.properties['fields']]
+    featurelayer_fields = [field['name'] for field in featurelayer.properties.fields]
     if index_column not in featurelayer_fields:
         raise RuntimeError(f'Index column {index_column} not found in feature layer fields {featurelayer_fields}')
 
 
+def rename_fields(dataframe, field_mapping):
+    """Rename fields based on field_mapping
+
+    Args:
+        dataframe (pd.DataFrame): Dataframe with columns to be renamed
+        field_mapping (dict): Mapping of existing field names to new names
+
+    Raises:
+        ValueError: If an existing name from field_mapping is not found in dataframe.columns
+
+    Returns:
+        pd.DataFrame: Dataframe with renamed fields
+    """
+
+    for original_name in field_mapping.keys():
+        if original_name not in dataframe.columns:
+            raise ValueError(f'Field {original_name} not found in dataframe.')
+
+    renamed_df = dataframe.rename(columns=field_mapping)
+
+    return renamed_df
+
+
 #: This isn't used anymore... but it feels like a shame to lose it.
 def build_sql_in_list(series):
     """Generate a properly formatted list to be a target for a SQL 'IN' clause
@@ -169,10 +193,10 @@ def check_field_set_to_unique(featurelayer, field_name):
         RuntimeError: If the field is not unique (or if it's indexed but not unique)
     """
 
-    fields = [field['fields'] for field in featurelayer.properties['indexes']]
+    fields = [field['fields'] for field in featurelayer.properties.indexes]
     if field_name not in fields:
         raise RuntimeError(f'{field_name} does not have a "unique constraint" set within the feature layer')
-    for field in featurelayer.properties['indexes']:
+    for field in featurelayer.properties.indexes:
         if field['fields'] == field_name:
             if not field['isUnique']:
                 raise RuntimeError(f'{field_name} does not have a "unique constraint" set within the feature layer')
@@ -305,3 +329,51 @@ def validate_api_key(api_key):
         module_logger.debug(error)
         # return 'Could not determine key validity; check your API key and/or network connection'
     return 'Could not determine key validity; check your API key and/or network connection'
+
+
+def authorize_pygsheets(credentials):
+    """Authenticate pygsheets using either a service file or google.auth.credentials.Credentials object.
+
+    Tries first to load credentials as a service account file; if this fails with a FileNotFoundError, tries to pass
+    credentials directly as a custom_credential.
+
+    Args:
+        credentials (str or google.auth.credentials.Credentials): Path to the service file OR credentials object
+            obtained from google.auth.default() within a cloud function.
+
+    Raises:
+        RuntimeError: If both authorization method attempts fail
+
+    Returns:
+        pygsheets.Client: Authorized pygsheets client
+    """
+
+    try:
+        return pygsheets.authorize(service_file=credentials)
+    except FileNotFoundError as err:
+        module_logger.debug(err)
+        module_logger.debug('Credentials file not found, trying as environment variable')
+    try:
+        return pygsheets.authorize(custom_credentials=credentials)
+    except Exception as err:
+        raise RuntimeError('Could not authenticate to Google API') from err
+
+
+def fix_numeric_empty_strings(feature_set, feature_layer_fields):
+    """Replace empty strings with None for numeric fields that allow nulls
+
+    Args:
+        feature_set (arcgis.features.FeatureSet): Feature set to clean
+        fields (Dict): fields from feature layer
+    """
+    fix_field_names = []
+    for field in feature_layer_fields:
+        if field['type'] in ['esriFieldTypeDouble', 'esriFieldTypeInteger'] and field['nullable']:
+            fix_field_names.append(field['name'])
+
+    for feature in feature_set.features:
+        for field_name in fix_field_names:
+            if feature.attributes[field_name] == '':
+                feature.attributes[field_name] = None
+
+    return feature_set
