@@ -1,4 +1,5 @@
 import logging
+from collections import namedtuple
 
 import numpy as np
 import pandas as pd
@@ -64,32 +65,28 @@ class TestCheckFieldsMatch:
 
     def test_check_fields_match_normal(self, mocker):
         mock_fl = mocker.Mock()
-        mock_fl.properties = {
-            'fields': [
-                {
-                    'name': 'Foo'
-                },
-                {
-                    'name': 'Bar'
-                },
-            ]
-        }
+        mock_fl.properties.fields = [
+            {
+                'name': 'Foo'
+            },
+            {
+                'name': 'Bar'
+            },
+        ]
         df = pd.DataFrame(columns=['Foo', 'Bar'])
 
         palletjack.utils.check_fields_match(mock_fl, df)
 
     def test_check_fields_match_raises_error_on_extra_new_field(self, mocker):
         mock_fl = mocker.Mock()
-        mock_fl.properties = {
-            'fields': [
-                {
-                    'name': 'Foo'
-                },
-                {
-                    'name': 'Bar'
-                },
-            ]
-        }
+        mock_fl.properties.fields = [
+            {
+                'name': 'Foo'
+            },
+            {
+                'name': 'Bar'
+            },
+        ]
         df = pd.DataFrame(columns=['Foo', 'Bar', 'Baz'])
 
         with pytest.raises(RuntimeError) as exc_info:
@@ -100,35 +97,31 @@ class TestCheckFieldsMatch:
 
     def test_check_fields_match_ignores_new_shape_field(self, mocker):
         mock_fl = mocker.Mock()
-        mock_fl.properties = {
-            'fields': [
-                {
-                    'name': 'Foo'
-                },
-                {
-                    'name': 'Bar'
-                },
-            ]
-        }
+        mock_fl.properties.fields = [
+            {
+                'name': 'Foo'
+            },
+            {
+                'name': 'Bar'
+            },
+        ]
         df = pd.DataFrame(columns=['Foo', 'Bar', 'SHAPE'])
 
         palletjack.utils.check_fields_match(mock_fl, df)
 
     def test_check_fields_match_warns_on_missing_new_field(self, mocker, caplog):
         mock_fl = mocker.Mock()
-        mock_fl.properties = {
-            'fields': [
-                {
-                    'name': 'Foo'
-                },
-                {
-                    'name': 'Bar'
-                },
-                {
-                    'name': 'Baz'
-                },
-            ]
-        }
+        mock_fl.properties.fields = [
+            {
+                'name': 'Foo'
+            },
+            {
+                'name': 'Bar'
+            },
+            {
+                'name': 'Baz'
+            },
+        ]
         df = pd.DataFrame(columns=['Foo', 'Bar'])
 
         palletjack.utils.check_fields_match(mock_fl, df)
@@ -241,31 +234,27 @@ class TestCheckIndexColumnInFL:
 
     def test_check_index_column_in_feature_layer_doesnt_raise_on_normal(self, mocker):
         fl_mock = mocker.Mock()
-        fl_mock.properties = {
-            'fields': [
-                {
-                    'name': 'Foo'
-                },
-                {
-                    'name': 'Bar'
-                },
-            ]
-        }
+        fl_mock.properties.fields = [
+            {
+                'name': 'Foo'
+            },
+            {
+                'name': 'Bar'
+            },
+        ]
 
         palletjack.utils.check_index_column_in_feature_layer(fl_mock, 'Foo')
 
     def test_check_index_column_in_feature_layer_raises_on_missing(self, mocker):
         fl_mock = mocker.Mock()
-        fl_mock.properties = {
-            'fields': [
-                {
-                    'name': 'Foo'
-                },
-                {
-                    'name': 'Bar'
-                },
-            ]
-        }
+        fl_mock.properties.fields = [
+            {
+                'name': 'Foo'
+            },
+            {
+                'name': 'Bar'
+            },
+        ]
 
         with pytest.raises(RuntimeError) as exc_info:
             palletjack.utils.check_index_column_in_feature_layer(fl_mock, 'Baz')
@@ -277,27 +266,23 @@ class TestCheckFieldUnique:
 
     def test_check_field_set_to_unique_doesnt_raise_on_normal(self, mocker):
         mock_fl = mocker.Mock()
-        mock_fl.properties = {
-            'indexes': [
-                {
-                    'fields': 'foo_field',
-                    'isUnique': True
-                },
-            ]
-        }
+        mock_fl.properties.indexes = [
+            {
+                'fields': 'foo_field',
+                'isUnique': True
+            },
+        ]
 
         palletjack.utils.check_field_set_to_unique(mock_fl, 'foo_field')
 
     def test_check_field_set_to_unique_raises_on_non_indexed_field(self, mocker):
         mock_fl = mocker.Mock()
-        mock_fl.properties = {
-            'indexes': [
-                {
-                    'fields': 'bar_field',
-                    'isUnique': True
-                },
-            ]
-        }
+        mock_fl.properties.indexes = [
+            {
+                'fields': 'bar_field',
+                'isUnique': True
+            },
+        ]
 
         with pytest.raises(RuntimeError) as exc_info:
             palletjack.utils.check_field_set_to_unique(mock_fl, 'foo_field')
@@ -306,14 +291,12 @@ class TestCheckFieldUnique:
 
     def test_check_field_set_to_unique_raises_on_non_unique_field(self, mocker):
         mock_fl = mocker.Mock()
-        mock_fl.properties = {
-            'indexes': [
-                {
-                    'fields': 'foo_field',
-                    'isUnique': False
-                },
-            ]
-        }
+        mock_fl.properties.indexes = [
+            {
+                'fields': 'foo_field',
+                'isUnique': False
+            },
+        ]
 
         with pytest.raises(RuntimeError) as exc_info:
             palletjack.utils.check_field_set_to_unique(mock_fl, 'foo_field')
@@ -634,3 +617,128 @@ class TestValidateAPIKey:
         assert check == 'Could not determine key validity; check your API key and/or network connection'
         assert palletjack.utils.requests.get.call_count == 4
         assert 'Random Error' in caplog.messages
+
+
+class TestFieldRenaming:
+
+    def test_rename_fields_renames_all_fields(self):
+        parcels_df = pd.DataFrame({
+            'account_no': [1, 2, 3],
+            'type': ['sf', 'mf', 'condo'],
+        })
+
+        field_mapping = {
+            'account_no': 'PARCEL_ID',
+            'type': 'class',
+        }
+
+        renamed_df = palletjack.utils.rename_fields(parcels_df, field_mapping)
+
+        assert list(renamed_df.columns) == ['PARCEL_ID', 'class']
+
+    def test_rename_fields_renames_some_fields(self):
+        parcels_df = pd.DataFrame({
+            'account_no': [1, 2, 3],
+            'class': ['sf', 'mf', 'condo'],
+        })
+
+        field_mapping = {
+            'account_no': 'PARCEL_ID',
+        }
+
+        renamed_df = palletjack.utils.rename_fields(parcels_df, field_mapping)
+
+        assert list(renamed_df.columns) == ['PARCEL_ID', 'class']
+
+    def test_rename_fields_raises_exception_for_missing_field(self):
+        parcels_df = pd.DataFrame({
+            'account_no': [1, 2, 3],
+            'type': ['sf', 'mf', 'condo'],
+        })
+
+        field_mapping = {
+            'account_no': 'PARCEL_ID',
+            'TYPE': 'class',
+        }
+
+        with pytest.raises(ValueError) as exception_info:
+            renamed_df = palletjack.utils.rename_fields(parcels_df, field_mapping)
+
+            assert 'Field TYPE not found in dataframe.' in str(exception_info)
+
+
+class TestAuthorization:
+
+    def test_authorize_pygsheets_auths_from_file(self, mocker):
+        pygsheets_mock = mocker.patch.object(palletjack.utils, 'pygsheets')
+        pygsheets_mock.authorize.return_value = 'authed'
+
+        client = palletjack.utils.authorize_pygsheets('file')
+
+        assert pygsheets_mock.authorize.called_once_with('file')
+        assert client == 'authed'
+
+    def test_authorize_pygsheets_auths_from_custom_credentials(self, mocker, caplog):
+        pygsheets_mock = mocker.patch.object(palletjack.utils, 'pygsheets')
+        pygsheets_mock.authorize.side_effect = [FileNotFoundError, 'authed']
+
+        caplog.set_level(logging.DEBUG, logger='palletjack.utils')
+        caplog.clear()
+
+        client = palletjack.utils.authorize_pygsheets('credentials')
+
+        assert 'Credentials file not found, trying as environment variable' in [rec.message for rec in caplog.records]
+
+        assert pygsheets_mock.authorize.call_count == 2
+        assert pygsheets_mock.authorize.call_args_list == [
+            mocker.call(service_file='credentials'),
+            mocker.call(custom_credentials='credentials')
+        ]
+        assert client == 'authed'
+
+    def test_authorize_pygsheets_raises_after_failing_both(self, mocker, caplog):
+        pygsheets_mock = mocker.patch.object(palletjack.utils, 'pygsheets')
+        pygsheets_mock.authorize.side_effect = [FileNotFoundError, IOError]
+
+        caplog.set_level(logging.DEBUG, logger='palletjack.utils')
+        caplog.clear()
+
+        with pytest.raises(RuntimeError):
+            client = palletjack.utils.authorize_pygsheets('credentials')
+
+        assert 'Credentials file not found, trying as environment variable' in [rec.message for rec in caplog.records]
+
+        assert pygsheets_mock.authorize.call_count == 2
+        assert pygsheets_mock.authorize.call_args_list == [
+            mocker.call(service_file='credentials'),
+            mocker.call(custom_credentials='credentials')
+        ]
+
+
+class TestEmptyStringsAsNulls:
+
+    def test_converts_empty_strings_to_null(self):
+        FeatureSet = namedtuple('FeatureSet', ['features'])
+        Feature = namedtuple('Feature', ['attributes'])
+        feature_set = FeatureSet([Feature({
+            'a': 'foo',
+            'b': 'baz',
+        }), Feature({
+            'a': '',
+            'b': '',
+        })])
+        fields = [{
+            'name': 'a',
+            'type': 'esriFieldTypeInteger',
+            'nullable': False,
+        }, {
+            'name': 'b',
+            'type': 'esriFieldTypeInteger',
+            'nullable': True,
+        }]
+
+        fixed_feature_set = palletjack.utils.fix_numeric_empty_strings(feature_set, fields)
+        fixed_feature = fixed_feature_set.features[1]
+
+        assert fixed_feature.attributes['a'] == ''
+        assert fixed_feature.attributes['b'] is None
